#include <stdio.h>
#include <stdlib.h>
#include <windows.h> // Probably need to make this more flexible
#include <openssl/sha.h> // Using SHA256 as random oracle
#include <math.h>
#include <time.h>

#define RtlGenRandom SystemFunction036 // CSPRNG of windows

BOOLEAN NTAPI RtlGenRandom(PVOID RandomBuffer, ULONG RandomBufferLength); // Used to pull secure random bits on windows

//#define q 170141183460469231731687303715884105727
#define TWO_128_MINUS_1 (*(unsigned __int128*) "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff")
#define TWO_64_MINUS_1 (*(unsigned __int64*) "\xff\xff\xff\xff\xff\xff\xff\xff")
#define TWO_63_MINUS_1 (*(unsigned __int64*) "\xff\xff\xff\xff\xff\xff\xff\x7f")
#define MODULO (*(unsigned __int128*) "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f") // 2^127-1
#define SMALLMOD (*(unsigned __int128*) "\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00") // 2^32
#define HALFSMALLMOD (*(unsigned __int128*) "\x00\x00\x00\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00") // 2^31
#define SCALE (*(unsigned __int128*) "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x00") // (2^127-1)/2^32
#define HALFSCALE (*(unsigned __int128*) "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00") // (2^127-1)/2^33

unsigned __int128 modq_add(unsigned __int128 a, unsigned __int128 b) // Arithmetic addition modulo 2^127-1
{
	unsigned __int128 x = a + b;
	unsigned __int128 y = x >> 127;
	return (x & MODULO)+y;
}

unsigned __int128 modq_mult(unsigned __int128 a, unsigned __int128 b) // Arithmetic multiplication modulo 2^127-1
{
	unsigned __int128 a_low = (unsigned __int128)((unsigned __int64)a);
	unsigned __int128 a_high = (unsigned __int128)((unsigned __int64)(a>>64));
	unsigned __int128 b_low = (unsigned __int128)((unsigned __int64)b);
	unsigned __int128 b_high = (unsigned __int128)((unsigned __int64)(b>>64));
	unsigned __int128 x = 2 * a_high * b_high;
	unsigned __int128 s = a_low*b_high;
	unsigned __int128 t = a_high*b_low;
	unsigned __int128 y = modq_add(s>>63,t>>63);
	unsigned __int128 u = modq_add((s&TWO_63_MINUS_1)<<64,(t&TWO_63_MINUS_1)<<64);
	unsigned __int128 z = a_low * b_low;
	unsigned __int128 o = z >> 127;
	return modq_add(modq_add(y,u),modq_add(x,(z & MODULO)+o));
}


void random_double_vector(__float128 **out, size_t len) // Generates uniform floating point numbers between 0 and 1
{
  	unsigned __int128 *tmp = (unsigned __int128*) malloc(len * sizeof(unsigned __int128));
  	RtlGenRandom((PVOID) tmp,(ULONG) len*sizeof(unsigned __int128));
	for(size_t i = 0; i < len; ++i) {
	  	(*out)[i] = (__float128) (tmp[i]) / (__float128) TWO_128_MINUS_1;
	  	//(*out)[i] = (double)rand() / (double)RAND_MAX;
	  	//(*out)[i] = 0;
	}
	free(tmp);
}

void random_gaussian_vector(unsigned __int128 **out, __float128 stddev, size_t len) // Generates "len" numbers from a rounded gaussian distribution
{
	// always generate an even number of samples (because of Box-Muller transform)
	if (len % 2 != 0)
		len += 1;
	__float128 *U = (__float128*) malloc(len * sizeof(__float128));
	//RtlGenRandom((PVOID) U,(ULONG) len*sizeof(__float128));
	random_double_vector(&U, len);
	for (size_t i = 0; i < len-1; i+=2) {
	  U[i] = sqrt(-2. * log(U[i]));
	  U[i+1] = 2. * M_PI * U[i+1];
	  //(*out)[i] = (unsigned __int128)((__int128)(stddev * U[i] * cos(U[i+1])));
	  //(*out)[i+1] = (unsigned __int128)((__int128)(stddev * U[i] * sin(U[i+1])));
	  (*out)[i] = (__int128)(stddev * U[i] * cos(U[i+1]));
	  (*out)[i+1] = (__int128)(stddev * U[i] * sin(U[i+1]));
	}
	free(U);
}

void print128(unsigned __int128 a){ // mostly used for debugging - decomposes a 128bits number into 2 64bits numbers and prints them as unsigned integers
	unsigned __int64 low;
	unsigned __int64 high;
	low = (unsigned __int64)a;
	high = (unsigned __int64)(a>>64);

	printf("%llu;%llu\n",high,low);
	return;
}

void prints128(unsigned __int128 a) // mostly used for debugging - decomposes a 128bits number into 2 64bits numbers and prints them as signed integers
{
	__int64 low;
	__int64 high;
	low = (__int64)a;
	high = (__int64)(a>>64);

	printf("%lld;%lld\n",high,low);
	return;
}

typedef uint32_t lwe_ct_t;// type of ciphertexts

typedef unsigned __int128* lwe_sk_t; // type of user secret keys

typedef lwe_sk_t* lwe_msk_t; // type of master secret keys

void lwe_sk_init(lwe_sk_t *sk, size_t sizeS, size_t sizeT) // used to allocate memory for a user secret key
{
	*sk = (lwe_sk_t) malloc ((sizeS+sizeT) * sizeof(unsigned __int128));
}

void lwe_sk_clear(lwe_sk_t sk) // used to free memory used by a user secret key
{
	free(sk);
}

void lwe_msk_init(lwe_msk_t *msk, size_t sizeS, size_t sizeT, size_t n) // used to allocate memory for a master secret key
{
	*msk = (lwe_msk_t) malloc (n * sizeof(lwe_sk_t));
	for(size_t i=0; i < n; ++i)
		lwe_sk_init((*msk)+i, sizeS, sizeT);
}

void lwe_msk_clear(lwe_msk_t msk, size_t n) // used to free memory used by a master secret key
{
	for(size_t i=0; i < n; ++i)
		lwe_sk_clear(msk[i]);	
	free(msk);
}

void lwe_keygen(lwe_sk_t *sk, __float128 stddev, size_t sizeS, size_t sizeT) // used to generate the i-th part of a master secret key
{
	RtlGenRandom((PVOID) (*sk),(ULONG) sizeS*sizeof(unsigned __int128));
	for (size_t i=0; i < sizeS; ++i)
		(*sk)[i] = (*sk)[i] >> 1;
	unsigned __int128 *t = (*sk)+sizeS;
	random_gaussian_vector(&t, stddev, sizeT);
}

void lwe_msk_keygen(lwe_msk_t *msk, __float128 stddev, size_t sizeS, size_t sizeT, size_t n) // used to generate a master secret key
{
	for (size_t i=0; i < n; ++i)
		lwe_keygen(*msk+i, stddev, sizeS, sizeT);
}

void lwe_keyder(lwe_sk_t *sky, lwe_msk_t msk, unsigned __int128* y, size_t sizeS, size_t sizeT, size_t n){ // used to generate a secret key associated to a vector y
	for(size_t j=0; j < (sizeS + sizeT); ++j)
		(*sky)[j] = 0;
	for(size_t i=0; i < n; ++i)
	{
		for(size_t j=0; j < (sizeS + sizeT); ++j)
		{
			(*sky)[j] = modq_add((*sky)[j],modq_mult(msk[i][j],y[i]));
		}
	}
}

void dot_product(unsigned __int128 *out, unsigned __int128 *a, unsigned __int128 *b, int len) // computes the dot product of two 128bits integers vectors
{
	*out = 0;
	for(int i = 0; i < len; ++i) {
		*out = modq_add(*out,modq_mult(a[i],b[i]));
	}
}

void lwe_encrypt(lwe_ct_t *ct, lwe_sk_t sk, int K, int x, unsigned char* tau, size_t size) // used to generate a partial ciphertext encrypting x for tag tau with secret key sk
{
	unsigned __int128 tmp;
	unsigned __int128* H;
	if (size % 2)
		H = malloc((size + 1) * sizeof(unsigned __int128));
	else
		H = malloc(size * sizeof(unsigned __int128));
	char* ctr = malloc(sizeof(size));
	SHA256_CTX h;
	for (size_t i=0; i < size; i+=2)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final((unsigned char*)(H+i), &h);
		H[i] = H[i] >> 1;
		H[i+1] = H[i+1] >> 1;
	}
	dot_product(&tmp, sk, H, size);
	tmp+=x*(MODULO/K);
	tmp=(tmp+HALFSCALE)/SCALE;
	(*ct) = (lwe_ct_t) tmp;
	free(H);
	free(ctr);
}

void lwe_decrypt(unsigned __int128 *m, lwe_sk_t sk, int K, lwe_ct_t* ct, unsigned __int128* y, unsigned char* tau, size_t size, size_t n) // used to decrypt a vector of ciphertexts with a functional secret key associated to vector y
{
	unsigned __int128 tmp;
	unsigned __int128* H;
	if (size % 2)
		H = malloc((size + 1) * sizeof(unsigned __int128));
	else
		H = malloc(size * sizeof(unsigned __int128));
	char* ctr = malloc(sizeof(size));
	SHA256_CTX h;
	for (size_t i=0; i < size; i+=2)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final((unsigned char*)(H+i), &h);
		H[i] = H[i] >> 1;
		H[i+1] = H[i+1] >> 1;
	}
	dot_product(&tmp, sk, H, size);
	tmp=(tmp+HALFSCALE)/SCALE;
	unsigned __int128 mu = 0;
	for (size_t i=0; i < n; ++i)
		mu += ct[i] * y[i];
	mu -= (lwe_ct_t) tmp;
	/*printf("here there are 4:\n");
	print128(mu);
	print128(mu*K);
	print128(mu*K+HALFSMALLMOD);
	print128(SMALLMOD);*/
	(*m) = ((mu*K+HALFSMALLMOD)/SMALLMOD)%K;
	free(H);
	free(ctr);
}

int main(){
	size_t sizeS = 5000;
	size_t sizeT = 500;
	size_t n = 100;
	size_t K = 1048576;
	__float128 stddev = 1000;

	struct timespec start, end;
	double deltat;

	lwe_msk_t msk;
	lwe_msk_init(&msk, sizeS, sizeT, n);

	clock_gettime(CLOCK_REALTIME, &start);
	//for (size_t i=0; i < 1000; ++i)
	lwe_msk_keygen(&msk, stddev, sizeS, sizeT, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Master secret key generation : %f s\n",deltat);

	/*for (int i = 0; i < sizeS; ++i)
	print128(msk[0][i]);
	for (int i = sizeS; i < sizeS+sizeT; ++i)
	prints128(msk[0][i]);*/

/*	print128(modq_add(21,modq_mult(MODULO-4,5)));
	print128(modq_add(21,MODULO-20));
	
	
	print128(SCALE-2*HALFSCALE);
	print128(HALFSCALE);*/

	unsigned __int128 x[n];
	unsigned __int128 y[n];
	for (size_t i=0; i < n; ++i)
	{
		x[i] = i;
		y[i] = i;
	}
	
	lwe_sk_t sky;
	lwe_sk_init(&sky, sizeS, sizeT);

	clock_gettime(CLOCK_REALTIME, &start);
	lwe_keyder(&sky, msk, y, sizeS, sizeT, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("User secret key derivation : %f s\n",deltat);

	lwe_ct_t ctx[n];

	clock_gettime(CLOCK_REALTIME, &start);
	for (size_t i=0; i < n; ++i)
		lwe_encrypt(ctx+i, msk[i], K, x[i], "a", sizeS+sizeT);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_nsec-start.tv_nsec)/1000000000.0/n;
	printf("Encryption : %f s\n",deltat);

	unsigned __int128 m;

	clock_gettime(CLOCK_REALTIME, &start);
	lwe_decrypt(&m, sky, K, ctx, y, "a", sizeS+sizeT, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Decryption : %f s\n",deltat);


	printf("decrypted\n");
	print128(m);
	printf("that was the value\n");

	unsigned __int128 result = 0;
	for (size_t i=0; i < n; ++i)
		result += x[i]*y[i];

	printf("expected was\n");
	print128(result);

	lwe_sk_clear(sky);
	lwe_msk_clear(msk, n);

	return 0;
}